from pyway.helpers import Utils


class Migration():
    def __init__(self, version, extension, name, checksum, apply_timestamp):
        self.version = version
        self.extension = extension
        self.name = name
        self.checksum = checksum
        self.apply_timestamp = apply_timestamp

    @classmethod
    def from_name(cls, name, path, **kwargs):
        version = kwargs.get('version', Utils.get_version_from_name(name))
        extension = kwargs.get('extension', Utils.get_extension_from_name(name))
        checksum = kwargs.get('checksum', Utils.load_checksum_from_name(name, path))
        apply_timestamp = kwargs.get('apply_timestamp')
        return cls(version, extension, name, checksum, apply_timestamp)

    @classmethod
    def from_list(cls, list_):
        return [cls(m.version, m.extension, m.name, m.checksum, m.apply_timestamp) for m in list_]

    @classmethod
    def from_list_names(cls, list_):
        return [cls.from_name(m.name) for m in list_]



import pytest
import os
from pyway.helpers import Utils


@pytest.mark.helpers_test
def test_get_local_files():
    files = Utils.get_local_files(os.path.join('tests', 'data'))
    assert len(files) == 3



import os
import sys

from pyway.log import logger
from pyway.migration import Migration
from pyway.dbms.database import factory
from pyway.helpers import Utils
from pyway.errors import VALID_NAME_ERROR


class Import():

    def __init__(self, args):
        self._db = factory(args.database_type)(args)
        self.migration_dir = args.database_migration_dir
        self.schema_file = args.schema_file
        self.args = args

    def run(self):
        if not self.schema_file:
            logger.error("Error, must specify --schema-file with import")
            sys.exit(1)

        if not os.path.exists(os.path.join(os.getcwd(), self.migration_dir, self.schema_file)):
            logger.error(f"Error, schema file '{self.migration_dir}/{self.schema_file}' does not exist!")
            sys.exit(1)

        if not Utils.is_file_name_valid(self.schema_file):
            logger.error(VALID_NAME_ERROR % (self.schema_file, Utils.expected_pattern()))
            sys.exit(1)

        # File exists, import it
        migration = Migration.from_name(self.schema_file, self.migration_dir)
        self._db.upgrade_version(migration)
        logger.info(f"{migration.name} Imported")



DIFF_NAME_ERROR = "ERROR: Local file name [%s] with diff name of the database [%s]"
OUT_OF_DATE_ERROR = "ERROR: Out of date - Local file missing [%s]"
DIFF_CHECKSUM_ERROR = "ERROR: Local file [%s] with diff script (%s) of the database (%s)"
VALID_NAME_ERROR = "ERROR: Local file [%s] has invalid format name - expected: %s"
DIRECTORY_NOT_FOUND = "ERROR: directory not found ('%s')"
MIGRATIONS_NOT_FOUND = "ERROR: no local migration files found in (%s) folder"
MIGRATIONS_NOT_STARTED = "ERROR: no migrations applied yet, no validation necessary."



import os
from pyway.log import logger
from pyway.helpers import Utils
from pyway.migration import Migration
from pyway.dbms.database import factory
from pyway.errors import MIGRATIONS_NOT_FOUND


class Migrate():

    def __init__(self, args):
        self._db = factory(args.database_type)(args)
        self.migration_dir = args.database_migration_dir
        self.args = args

    def run(self):
        migrations_to_be_executed = self._get_migration_files_to_be_executed()
        if not migrations_to_be_executed:
            logger.info("Nothing to do")
            return

        for migration in migrations_to_be_executed:
            logger.info(f"Migrating --> {migration.name}")
            try:
                with open(os.path.join(os.getcwd(),
                          self.migration_dir, migration.name), "r", encoding='utf-8') as sqlfile:
                    self._db.execute(sqlfile.read())
                self._db.upgrade_version(migration)
                logger.success(f"{migration.name} SUCCESS")
            except Exception as error:
                logger.error(error)

    def _get_migration_files_to_be_executed(self):
        all_local_migrations = self._get_all_local_migrations()
        all_db_migrations = Migration.from_list(self._db.get_all_schema_migrations())

        if all_db_migrations and not all_local_migrations:
            logger.error(MIGRATIONS_NOT_FOUND % self.migration_dir)
        return Utils.subtract(all_local_migrations, all_db_migrations)

    def _get_all_local_migrations(self):
        local_files = Utils.get_local_files(self.migration_dir)
        if not local_files:
            return []
        migrations = [Migration.from_name(local_file, self.migration_dir) for local_file in local_files]
        return Utils.sort_migrations_list(migrations)



import sys
from pyway.log import logger
from pyway.helpers import Utils
from pyway.dbms.database import factory
from pyway.migration import Migration
from pyway.errors import (OUT_OF_DATE_ERROR, DIFF_NAME_ERROR, DIFF_CHECKSUM_ERROR,
                          VALID_NAME_ERROR, MIGRATIONS_NOT_FOUND, MIGRATIONS_NOT_STARTED)


class Validate():
    def __init__(self, args):
        self._db = factory(args.database_type)(args)
        self.migration_dir = args.database_migration_dir
        self.args = args

    def run(self):
        local_migrations = self._get_all_local_migrations()
        db_migrations = self._db.get_all_schema_migrations()

        if not db_migrations:
            logger.error(MIGRATIONS_NOT_STARTED)
            sys.exit(1)

        if db_migrations and not local_migrations:
            logger.error(MIGRATIONS_NOT_FOUND % self.migration_dir)

        if local_migrations:
            local_migrations_map = Utils.create_map_from_list("version", local_migrations)

            for db_migration in db_migrations:
                logger.info(f"Validating --> {db_migration.name}")
                local_migration = local_migrations_map.get(db_migration.version)

                if not self._out_of_date(local_migration):
                    logger.error(OUT_OF_DATE_ERROR % db_migration.name)
                elif not self._name_format(local_migration.name):
                    logger.error(VALID_NAME_ERROR % (local_migration.name, Utils.expected_pattern()))
                elif not self._diff_names(local_migration, db_migration):
                    logger.error(DIFF_NAME_ERROR % (local_migration.name, db_migration.name))
                elif not self._diff_checksum(local_migration, db_migration):
                    logger.error(DIFF_CHECKSUM_ERROR % (local_migration.name,
                                                        local_migration.checksum,
                                                        db_migration.checksum))
                else:
                    logger.success(f"{db_migration.name} VALID")

    def _out_of_date(self, local_migration):
        return bool(local_migration is not None)

    def _diff_names(self, local_migration, db_migration):
        return bool(local_migration.name == db_migration.name)

    def _diff_checksum(self, local_migration, db_migration):
        return bool(local_migration.checksum == db_migration.checksum)

    def _name_format(self, name):
        return bool(Utils.is_file_name_valid(name))

    def _get_all_local_migrations(self):
        local_files = Utils.get_local_files(self.migration_dir)
        if not local_files:
            return []
        migrations = [Migration.from_name(local_file, self.migration_dir) for local_file in local_files]
        return Utils.sort_migrations_list(migrations)



from tabulate import tabulate

from pyway.helpers import Utils
from pyway.log import bcolors
from pyway.migration import Migration
from pyway.dbms.database import factory


class Info():

    def __init__(self, config):
        self.migration_dir = config.database_migration_dir
        self._db = factory(config.database_type)(config)
        self.headers = ["version", "extension", "name", "checksum", "apply_timestamp"]
        self.tablefmt = "psql"
        self.config = config

    def run(self):
        return tabulate(Utils.flatten_migrations(self.get_table_info()), headers="keys", tablefmt=self.tablefmt)

    def get_table_info(self):
        db_migrations = self._db.get_all_schema_migrations()
        local_migrations = self.get_new_local_migrations(db_migrations, self.migration_dir)
        return db_migrations + local_migrations

    def get_new_local_migrations(self, db_migrations, migration_dir):
        local_files = Utils.get_local_files(migration_dir)
        if not local_files:
            return []

        new_local_migrations = [self.structure_migration(local_file) for local_file in local_files
                                if local_file not in [db_migration.name for db_migration in db_migrations]]
        return Utils.sort_migrations_list(new_local_migrations)

    def structure_migration(self, name):
        checksum = "%snew%s" % (bcolors.OKGREEN, bcolors.OKBLUE)
        apply_timestamp = "%snew%s" % (bcolors.OKGREEN, bcolors.OKBLUE)
        return Migration.from_name(name, self.migration_dir, checksum=checksum, apply_timestamp=apply_timestamp)



import pytest
from pyway.settings import ConfigFile


# Make sure config options exists and check some defaults
@pytest.mark.settings_test
def test_settings_database_migration_dir():
    config = ConfigFile()
    assert config.database_migration_dir == 'resources'


@pytest.mark.settings_test
def test_settings_database_table():
    config = ConfigFile()
    assert config.database_table == 'public.pyway'


@pytest.mark.settings_test
def test_settings_database_type():
    config = ConfigFile()
    assert config.database_type == 'postgres'


@pytest.mark.settings_test
def test_settings_database_host():
    config = ConfigFile()
    assert config.database_host == 'localhost'


@pytest.mark.settings_test
def test_settings_database_port():
    config = ConfigFile()
    assert config.database_port == '5432'


@pytest.mark.settings_test
def test_settings_database_name():
    config = ConfigFile()
    assert config.database_name == 'postgres'


@pytest.mark.settings_test
def test_settings_database_username():
    config = ConfigFile()
    assert config.database_username == 'postgres'


@pytest.mark.settings_test
def test_settings_database_password():
    config = ConfigFile()
    assert config.database_password == 'password'


@pytest.mark.settings_test
def test_settings_schema_file():
    config = ConfigFile()
    assert config.schema_file is None


@pytest.mark.settings_test
def test_settings_config():
    config = ConfigFile()
    assert config.config == '.pyway.conf'


@pytest.mark.settings_test
def test_settings_version():
    config = ConfigFile()
    assert config.version is False


@pytest.mark.settings_test
def test_settings_cmd():
    config = ConfigFile()
    assert config.cmd is None



import pytest
import os
from strip_ansi import strip_ansi
from pyway.info import Info
# from pyway.migrate import Migrate
# from pyway.validate import Validate
# from pyway.import_ import Import
# from pyway.settings import Settings
from pyway.settings import ConfigFile
from mysqld_integration_test import Mysqld
import mysql.connector

INFO_TABLE = """+-----------+-------------+-------------------+------------+-------------------+
|   version | extension   | name              | checksum   | apply_timestamp   |
|-----------+-------------+-------------------+------------+-------------------|
|      1.01 | SQL         | V01_01__test1.sql | new        | new               |
|      1.02 | SQL         | V01_02__test2.sql | new        | new               |
|      1.03 | SQL         | V01_03__test3.sql | new        | new               |
+-----------+-------------+-------------------+------------+-------------------+"""


def execute_query(mysqld, query):
    cnx = mysql.connector.connect(user=mysqld.username,
                                  password=mysqld.password,
                                  host=mysqld.host,
                                  port=mysqld.port, database='test')
    cursor = cnx.cursor()
    cursor.execute(query)
    cnx.commit()
    cursor.close()
    cnx.close()


def select_query(mysqld, query):
    cnx = mysql.connector.connect(user=mysqld.username,
                                  password=mysqld.password,
                                  host=mysqld.host,
                                  port=mysqld.port, database='test')
    cursor = cnx.cursor()
    cursor.execute(query)
    for _result in cursor:
        result = _result
    cursor.close()
    cnx.close()

    return result[0]


@pytest.fixture
def mysqld_connect(autouse=True):
    mysqld = Mysqld()
    return mysqld.run()


@pytest.mark.integration_test
def test_pyway_table_creation(mysqld_connect):
    config = ConfigFile()
    config.database_type = "mysql"
    config.database_host = mysqld_connect.host
    config.database_username = mysqld_connect.username
    config.database_password = mysqld_connect.password
    config.database_port = mysqld_connect.port
    config.database_name = 'test'
    config.database_table = 'pyway'
    config.database_migration_dir = os.path.join('tests', 'data')
    tbl = Info(config).run()

    assert strip_ansi(tbl) == INFO_TABLE



import os
import argparse
import yaml

# Pyway consts
SQL_MIGRATION_PREFIX = os.environ.get('PYWAY_SQL_MIGRATION_PREFIX', 'V')
SQL_MIGRATION_SEPARATOR = os.environ.get('PYWAY_SQL_MIGRATION_SEPARATOR', '__')
SQL_MIGRATION_SUFFIXES = os.environ.get('PYWAY_SQL_MIGRATION_SUFFIXES', '.sql')
ARGS = ['database_migration_dir', 'database_table', 'database_type', 'database_host',
        'database_port', 'database_name', 'database_username', 'database_password',
        'schema_file', 'config', 'version', 'cmd']


def parse_args(config, args):
    for arg in ARGS:
        if getattr(args, arg):
            setattr(config, arg, getattr(args, arg))
    return config


class Settings():

    def __init__(self, args):
        self.args = args

    def parse_arguments(config):
        parser = argparse.ArgumentParser()
        parser.add_argument("--database-migration-dir", help="Database migration directory")
        parser.add_argument("--database-table", help="Database table that stores pyway metadata")
        parser.add_argument("--database-type", help="Database type [postgres|mysql]")
        parser.add_argument("--database-host", help="Database host")
        parser.add_argument("--database-port", help="Database port")
        parser.add_argument("--database-name", help="Database name")
        parser.add_argument("--database-username", help="Database username")
        parser.add_argument("--database-password", help="Database password")

        parser.add_argument("--schema-file", help="Schema file for import")
        parser.add_argument("-c", "--config", help="Config file")
        parser.add_argument("-v", "--version", help="Version", action='store_true')
        parser.add_argument("cmd", nargs="?", help="info|validate|migrate|import")

        config = parse_args(config, parser.parse_args())
        return (config, parser)

    def parse_config_file(config):
        # See if there is a config file
        if os.path.exists(config.config):
            with open(config.config, "r", encoding='utf-8') as ymlfile:
                cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)

            # Merge config together with args
            for c in cfg:
                setattr(config, c, cfg[c])

        return config


class ConfigFile():
    def __init__(self):
        self.database_migration_dir = os.environ.get('PYWAY_DATABASE_MIGRATION_DIR', 'resources')
        self.database_table = os.environ.get('PYWAY_TABLE', 'public.pyway')
        self.database_type = os.environ.get('PYWAY_TYPE', 'postgres')
        self.database_host = os.environ.get('PYWAY_DATABASE_HOST', 'localhost')
        self.database_port = os.environ.get('PYWAY_DATABASE_PORT', '5432')
        self.database_name = os.environ.get('PYWAY_DATABASE_NAME', 'postgres')
        self.database_username = os.environ.get('PYWAY_DATABASE_USERNAME', 'postgres')
        self.database_password = os.environ.get('PYWAY_DATABASE_PASSWORD', 'password')
        self.schema_file = None
        self.config = os.environ.get('PYWAY_CONFIG_FILE', '.pyway.conf')
        self.version = False
        self.cmd = None



import sys
import logging


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class _Log():
    def __init__(self):
        self.logger = logging.getLogger('pyway')
        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        self.logger.setLevel(logging.INFO)

    def debug(self, msg):
        if self.logger:
            self.logger.debug(msg)

    def info(self, msg):
        if self.logger:
            self.logger.info(self._colored(msg, bcolors.OKBLUE))

    def error(self, msg):
        if self.logger:
            self.logger.error(self._colored(msg, bcolors.FAIL))

    def warn(self, msg):
        if self.logger:
            self.logger.warn(self._colored(msg, bcolors.WARNING))

    def _colored(self, msg, color):
        return f"{color}{msg}{bcolors.ENDC}"

    def success(self, msg):
        if self.logger:
            self.logger.info(self._colored(msg, bcolors.OKGREEN))


logger = _Log()



import sys

from pyway.settings import Settings
from pyway.settings import ConfigFile
from pyway.info import Info
from pyway.log import logger
from pyway.migrate import Migrate
from pyway.validate import Validate
from pyway.import_ import Import
from pyway.version import __version__


def migrate(config):
    logger.info('Starting migration process...')
    Migrate(config).run()
    logger.info('Migration completed.')


def validate(config):
    logger.info('Starting validation process')
    Validate(config).run()
    logger.info('Validation completed.')


def info(config):
    logger.info('Gathering info...')
    tbl = Info(config).run()
    logger.info(tbl)
    print()


def import_(config):
    logger.info("Importing schema...")
    Import(config).run()


def cli():
    logger.info(f"PyWay {__version__}")

    config = ConfigFile()
    config = Settings.parse_config_file(config)
    (config, parser) = Settings.parse_arguments(config)

    # Display version if it exists
    # TODO: FIX
    if config.version:
        print(f"Version: {__version__}")
        sys.exit(1)

    # If no arg is specified, show help
    if not config.cmd:
        # TODO: figure out how to print help
        parser.print_help()
        sys.exit(1)

    if config.cmd == "info":
        info(config)
    elif config.cmd == "validate":
        validate(config)
    elif config.cmd == "migrate":
        migrate(config)
    elif config.cmd == "import":
        import_(config)
    else:
        logger.error(f"Command '{config.cmd}' not recognized, exiting!")
        sys.exit(1)


if __name__ == '__main__':
    cli()



import os
import re
import zlib

from pyway import settings
from pyway.log import logger
from pyway.errors import VALID_NAME_ERROR, DIRECTORY_NOT_FOUND, OUT_OF_DATE_ERROR


class Utils():

    @staticmethod
    def subtract(list_a, list_b):
        result = []
        if list_a and list_b:
            checksum_list_b = [b.checksum for b in list_b]
            result = [a for a in list_a if a.checksum not in checksum_list_b]
        elif list_a and not list_b:
            # List B is empty (usually from a new install)
            return list_a
        return result

    @staticmethod
    def expected_pattern():
        return f'{settings.SQL_MIGRATION_PREFIX}{{major}}_{{minor}}{settings.SQL_MIGRATION_SEPARATOR}' \
                '{{description}}{settings.SQL_MIGRATION_SUFFIXES}'

    @staticmethod
    def is_file_name_valid(name):
        _pattern = r"^%s\d+_\d{2}%s\w+%s$" % \
            (settings.SQL_MIGRATION_PREFIX, settings.SQL_MIGRATION_SEPARATOR, settings.SQL_MIGRATION_SUFFIXES)
        return re.match(_pattern, name, re.IGNORECASE) is not None

    @staticmethod
    def sort_migrations_list(migrations):
        return sorted(migrations, key=lambda x: [x.get("version"), x.get("name")] if isinstance(x, dict) else
                                                [x.version, x.name], reverse=False)

    @staticmethod
    def flatten_migrations(migrations):
        migration_list = []
        for migration in migrations:
            migration_list.append({'version': migration.version, 'extension': migration.extension,
                                   'name': migration.name, 'checksum': migration.checksum,
                                   'apply_timestamp': migration.apply_timestamp})
        return migration_list

    @staticmethod
    def get_version_from_name(name):
        try:
            return re.findall(r"\d+_\d{2}", name)[0].replace('_', '.')
        except IndexError:
            logger.error(VALID_NAME_ERROR % (name, Utils.expected_pattern()))
            return None

    @staticmethod
    def get_extension_from_name(name):
        return name.split('.')[1].upper()

    @staticmethod
    def load_checksum_from_name(name, path):
        fullname = os.path.join(os.getcwd(), path, name)
        prev = 0
        try:
            for line in open(fullname, "rb"):
                prev = zlib.crc32(line, prev)
            return "%X" % (prev & 0xFFFFFFFF)
        except FileNotFoundError:
            logger.error(OUT_OF_DATE_ERROR % fullname.split("/")[-1])
            return None

    @staticmethod
    def basepath(d):
        return os.path.join(os.getcwd(), d)

    @staticmethod
    def get_min_version_from_local_migrations():
        return min([int(Utils.get_version_from_name(file.name)) for file in os.listdir(Utils.basepath())])

    @staticmethod
    def get_local_files(d):
        path = Utils.basepath(d)
        dir_list = None
        try:
            dir_list = os.listdir(path)
        except OSError:
            logger.error(DIRECTORY_NOT_FOUND % path)
        return dir_list

    @staticmethod
    def create_map_from_list(key, list_):
        return {lst.__dict__[key]: lst for lst in list_}



import pytest
from pyway.version import __version__


@pytest.mark.version_test
def test_version_isnotnone():
    assert __version__ is not None


@pytest.mark.version_test
def test_version_semver():
    assert len(__version__.split('.')) == 3



import psycopg2
from pyway.migration import Migration


CREATE_VERSION_MIGRATIONS = "create table if not exists %s ("\
    "installed_rank serial PRIMARY KEY,"\
    "version varchar(20) NOT NULL,"\
    "extension varchar(20) NOT NULL,"\
    "name varchar(125) NOT NULL,"\
    "checksum varchar(25) NOT NULL,"\
    "apply_timestamp timestamp DEFAULT NOW()"\
    ");"
SELECT_FIELDS = ("version", "extension", "name", "checksum", "apply_timestamp")
ORDER_BY_FIELD_ASC = "installed_rank"
ORDER_BY_FIELD_DESC = "installed_rank desc"
INSERT_VERSION_MIGRATE = "insert into %s (version, extension, name, checksum) values ('%s', '%s', '%s', '%s');"


class Postgres():

    def __init__(self, args):
        self.args = args
        self.version_table = args.database_table
        self.create_version_table_if_not_exists()

    def connect(self):
        return psycopg2.connect(f"dbname={self.args.database_name} user={self.args.database_username} host={self.args.database_host} password={self.args.database_password} port={self.args.database_port}")  # noqa: E501

    def create_version_table_if_not_exists(self):
        self.execute(CREATE_VERSION_MIGRATIONS % self.version_table)

    def execute(self, script):
        conn = self.connect()
        cur = conn.cursor()
        cur.execute(script)
        conn.commit()

    def get_all_schema_migrations(self):
        cnx = self.connect()
        cursor = cnx.cursor()
        cursor.execute(f"SELECT {','.join(SELECT_FIELDS)} FROM {self.version_table} ORDER BY {ORDER_BY_FIELD_ASC}")
        migrations = []
        for row in cursor.fetchall():
            migrations.append(Migration(row[0], row[1], row[2], row[3], row[4]))
        cursor.close()
        cnx.close()
        return migrations

    def upgrade_version(self, migration):
        self.execute(INSERT_VERSION_MIGRATE % (self.version_table, migration.version,
                                               migration.extension, migration.name,
                                               migration.checksum))
